# -*- coding: utf-8 -*-

"""
the common functionalities
"""


def get_value_from_dict_safe(d, key, default=None):
    """
    get the value from dict
    args:
        d: {dict}
        key: {a hashable key, or a list of hashable key}
            if key is a list, then it can be assumed the d is a nested dict
        default: return value if the key is not reachable, default is None
    return:
        value
    """
    assert isinstance(d, dict), f"only supports dict input, {type(d)} is given"
    if isinstance(key, (list, tuple)):
        for _k in key[:-1]:
            if _k in d and isinstance(d[_k], dict):
                d = d[_k]
            else:
                return default
        key = key[-1]
    return d.get(key, default)


def set_value_to_dict_safe(d, key, value, append=False):
    """
    set the value to dict
    args:
        d: {dict}
        key: {a hashable key, or a list of hashable key}
            if key is a list, then it can be assumed the d is a nested dict
        value: value to be set
        append: if the value is appended to the list, default is False
    return:
        bool: if the value is succesfully set
    """
    assert isinstance(d, dict), f"only supports dict input, {type(d)} is given"
    if isinstance(key, (list, tuple)):
        for _k in key[:-1]:
            if _k in d:
                if isinstance(d[_k], dict):
                    d = d[_k]
                else:
                    return False
            else:
                d[_k] = dict()
                d = d[_k]
        key = key[-1]
    if append:
        if key not in d:
            d[key] = [value]
        elif isinstance(d[key], list):
            d[key].append(value)
        else:
            return False
    else:
        d[key] = value
    return True


def visualize_dict(d, indent="--", level=0):
    """
    print out the the dict strctures, unwraping list by 1-depth
    """
    prefix = indent * level
    if prefix:
        prefix = " |" + prefix
    if isinstance(d, (list, tuple)):
        print(prefix, f"[{type(d[0])} * {len(d)}]")
        return
    elif not isinstance(d, dict):
        print(prefix, f"{type(d)}")
        return

    for k, v in d.items():
        if isinstance(v, (dict, )):
            print(prefix, k)
            visualize_dict(v, indent=indent, level=level + 1)
        elif isinstance(v, (list, tuple)):
            print(prefix, k)
            if isinstance(v[0], (dict, )):
                for i in v:
                    visualize_dict(i, indent=indent, level=level + 1)
            else:
                visualize_dict(v, indent=indent, level=level + 1)
        else:
            print(prefix, k, v)


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# -*- coding: utf-8 -*-
"""
dicom related functionalities
"""
import os.path as osp
import numpy as np
import SimpleITK as sitk


DEFAULT_DICOM_TAG = {
    "patientID": "0010|0020",
    "studyUID": "0020|000d",
    "seriesUID": "0020|000e",
    "customUID": "0008|103e",
    "image_pixel_spacing": "0018|1164",
    "instance_number": "0020|0013",
    "manufacturer": "0008|0070",
    "body_part": "0018|0015",
    "body_part_thickness": "0018|11a0",
    "primary_angle": "0018|1510",
    "view": "0018|5101",
    "laterality": "0020|0062",
    "window_center": "0028|1050",
    "window_width": "0028|1051",
    "rescale_intercept": "0028|1052",
    "rescale_slope": "0028|1053",
    "patient_orientation": "0020|0020",
    "PresentationLUTShape": "2050|0020",
    "sop_instance_uid": "0008|0018"
}


def sitk_read_image(img_path, as_np=False):
    try:
        img = sitk.ReadImage(img_path)
        if as_np:
            img = sitk.GetArrayFromImage(img)
    except Exception:
        print(f"[Error] unable to load img_path {img_path}, "
              "perhaps its not standard format")
        return None
    return img


def get_image_info(img_path, info=None):
    """
    read dicom tags and return their values as dict
    args:
        img_path (str): the image path
        info (dict{tag_name->tag_position})
    return:
        info_dict:  the dicom tag values, default is 'None'
    """
    parsing_tags = DEFAULT_DICOM_TAG.copy()
    if info is not None:
        parsing_tags.update(info)
    info_dict = {tag: None for tag in parsing_tags}
    if isinstance(img_path, sitk.Image):
        img_itk = img_path
    elif isinstance(img_path, str):
        if not osp.exists(img_path):
            print("[Error]image_path does not exist")
            return info_dict
        img_itk = sitk_read_image(img_path)
        if img_itk is None:
            return info_dict
    for tag, meta_key in parsing_tags.items():
        try:
            info_dict[tag] = img_itk.GetMetaData(meta_key).strip(" \n")
        except Exception:
            info_dict[tag] = None
    return info_dict


def sitk_read_image_series(image_series, uid=None, verbose=False):
    """
    reading image series into a 3d image stack
    """
    reader = sitk.ImageSeriesReader()
    if isinstance(image_series, (list, set, tuple)):
        if not np.all([osp.exists(path) for path in image_series]):
            print(
                "[WARNING] some images are missing"
            )
    elif isinstance(image_series, str):
        if not osp.isdir(image_series):
            print("[ERROR] specified directory is not existed")
            return
        else:
            if uid is None:
                image_series = reader.GetGDCMSeriesFileNames(
                    image_series, loadSequences=True)
            else:
                image_series = reader.GetGDCMSeriesFileNames(
                    image_series, uid, loadSequences=True)
    try:
        if verbose:
            print(image_series)
        reader.SetFileNames(image_series)
        img_itk = reader.Execute()
    except Exception:
        img_itk = None
    return img_itk


def update_tags(img_path, update_dict):
    """
    update tags
    Args:
        img_path(str): path
        update_dict(dict{tag_key: value})
    """
    img = sitk.ReadImage(img_path)
    for key, value in update_dict.items():
        if key in DEFAULT_DICOM_TAG:
            key = DEFAULT_DICOM_TAG[key]
        img.SetMetaData(key, value)
    sitk.WriteImage(img, img_path)


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

'''
zqy's utility functions
'''
__version__ = "0.1.7"
__short_description__ = "zqy's utility functions."
__license__ = "MIT"
__author__ = "qianyizhang"
__author_email__ = "landy123007@hotmail.com"
__maintainer__ = "qianyizhang"
__maintainer_email__ = "landy123007@hotmail.com"
__github_username__ = "qianyizhang"
__github_url__ = "https://github.com/qianyizhang/zqy-utils"
__support__ = "https://github.com/qianyizhang/zqy-utils/issues"

from .debug import *    # noqa
from .dicom import *    # noqa
from .dict import *     # noqa
from .io import *       # noqa
from .plot import *     # noqa
from .point import *    # noqa
from .registry import * # noqa


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

import itertools

import numpy as np


def flat_nested_list(nested_list):
    return list(itertools.chain(*nested_list))


def get_bounding_box(edge_list, dim=2):
    """
    given a (nested) list of points, return its bounding box
    args:
        edge_list (list[points])
        dim (int): point dimension, default is 2
        if the list is empty, all values are defaulted to -1
    return:
        bounding_box: (2d np.array with shape 2xdim): [top_left, bot_right]
    """
    edge_list_np = np.array(flat_nested_list(edge_list))
    try:
        edge_list_np = edge_list_np.reshape(-1, dim)
        min_point = edge_list_np.min(axis=0)
        max_point = edge_list_np.max(axis=0)
        bounding_box = np.vstack((min_point, max_point))
    except Exception:
        print(edge_list)
        dtype = edge_list_np.dtype
        print(dtype)
        bounding_box = np.ones((2, dim), dtype=dtype) * -1
    return bounding_box


def psf(pts, kernel=0):
    """
    point spread function
    Args:
        pts (list[float]): K-dim point
        kenerl (int) : 0 = center, 1=8 points, 2=27 points
    Return:
        pts_list (tuple(1d array)): K array with N points

    Note: the N points count is using 3-d input as reference.
    """
    if kernel == 0:
        pts = np.array(pts).round().astype(int)
    elif kernel == 1:
        dim_list = [(int(pt), int(pt+1)) for pt in pts]
        pts = np.meshgrid(*tuple(dim_list))
    elif kernel == 2:
        neighbor_pts = np.meshgrid(*[(-1, 0, 1)]*3)
        pts = [(n+round(pt)).astype(int) for n, pt in zip(neighbor_pts, pts)]
    return tuple(pts)


def _test_psf():
    volume = np.zeros((10, 10, 10))
    pt = np.random.random(3)
    pt = pt + [3, 4, 5]

    volume[psf(pt, 0)] = 1
    assert volume.sum() == 1
    volume[psf(pt, 1)] = 2
    assert volume.sum() == 2 * 8
    volume[psf(pt, 2)] = 3
    assert volume.sum() == 3 * 27


def union_merge(merge_mat):
    """
    return group_indices based on merge_mat
    Args:
        merge_mat (NxN np.array or torch.Tensor): merging criteria
    Return:
        group_indices (list[indices])
    """
    N = len(merge_mat)
    if N == 0:
        return []
    else:
        item_id = np.arange(N)
        group_indices = [[i] for i in range(N)]
        for id1 in range(N):
            for id2 in range(N):
                if not merge_mat[id1, id2]:
                    continue
                min_id = min(item_id[id1], item_id[id2])
                for cur_id in (item_id[id1], item_id[id2]):
                    if cur_id == min_id:
                        continue
                    group_indices[min_id].extend(group_indices[cur_id])
                    group_indices[cur_id] = []
                    item_id[group_indices[min_id]] = min_id

    group_indices = [i for i in group_indices if i != []]
    return group_indices


def get_pts_merge_mat(pts_list, pts_list2=None, ratio=0.25, criteria="min"):
    """
    get the merge_mat for points list
    Args:
        pts_list (list[np.ndarray]):
            each item in pts_list is an array of points
        pts_list2 (list[np.ndarray]):
            secondary list, if is not given, assuming it's self merge
    Return:
        merge_mat (N x N np.ndarray): binary mat
    """

    if pts_list2 is None:
        replicate = True
        pts_list2 = pts_list
    else:
        replicate = False
    M = len(pts_list)
    N = len(pts_list2)

    merge_mat = np.ones((M, N))
    for i in range(M):
        for j in range(N):
            if i >= j and replicate:
                merge_mat[i, j] = merge_mat[j, i]
                continue
            pts1, pts2 = pts_list[i], pts_list2[j]
            num1, num2 = len(pts1), len(pts2)
            # get shared pts
            pts_all = np.concatenate([pts1, pts2])
            num_union = len(np.unique(pts_all, axis=0))
            divident = num1 + num2 - num_union
            if criteria == "min":
                divisor = min(num1, num2)
            elif criteria == "iou":
                divisor = num_union
            elif criteria == "self":
                divisor = num1
            elif criteria == "ref":
                divisor = num2
            else:
                raise NotImplementedError(f"unkown criteria {criteria}")
            merge_mat[i, j] = divident * 1.0 / divisor
    return merge_mat > ratio


def get_rounded_pts(pts_list, index_range, stride=1.0, as_unique=False):
    """
    given a list of points, cast them to int
    """
    start, end = index_range
    assert end >= start, f"invalid index_range {index_range}"
    pts = np.array(pts_list)
    pts = (pts[start:end + 1] / stride).round() * stride
    pts = pts.astype(int)
    if as_unique:
        return np.unique(pts, axis=0)
    else:
        return pts


def _test_union_merge():
    def has_match(pts):
        return (merge_fn(pts) - np.eye(len(pts))).max(0)
    pts = np.random.rand(10)

    def merge_fn(pts):
        return abs(pts[None] - pts[:, None]) < 0.1

    merge_mat = merge_fn(pts)
    indices_group = union_merge(merge_mat)
    valid = True
    for indices in indices_group:
        if len(indices) == 1:
            continue
        if not has_match(pts[indices]).all():
            print("pts within group is not connected to others")
            print(merge_mat, indices_group, indices)
            valid = False
            break

    for _ in range(10):
        indices = [np.random.choice(i) for i in indices_group]
        if has_match(pts[indices]).any():
            print("pts between groups is connected")
            print(merge_mat, indices_group, indices)
            valid = False
            break
    return valid


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# -*- coding: utf-8 -*-

import cv2
import numpy as np

DEFAULT_PALETTE = [3**7 - 1, 2**7 - 1, 5**9 - 1]


def get_colors(labels, palette=DEFAULT_PALETTE):
    """
    Simple function convert label to color
    Args:
        labels (int or list[int])
        palette ([R, G, B])
    Return:
        return (2d np.array): N x 3, with dtype=uint8
    """
    colors = np.array(labels).reshape(-1, 1) * palette
    colors = (colors % 255).astype("uint8")
    return colors


def img_to_uint8(img, img_max=None, img_min=None):
    img_float = img.astype(np.float32)
    if img_max is None:
        img_max = img_float.max()
    if img_min is None:
        img_min = img_float.min()
    img_float = (img_float - img_min) / (img_max - img_min) * 255.0
    return img_float.clip(0, 255).astype(np.uint8)


def overlay_bboxes(image, boxes, colors, line_thickness=5):
    image = np.array(image)
    if image.ndim == 2:
        image = image[..., None]
    assert image.ndim == 3, f"unknown image shape {image.shape}"
    if image.shape[0] in (1, 3):
        # chw -> hwc
        image = image.transpose(1, 2, 0)
    if image.shape[-1] == 1:
        image = image.repeat(3, -1)

    # image should be h *w* 3 now

    for box, color in zip(boxes, colors):
        box = box.round().astype(int)
        top_left, bottom_right = box[:2].tolist(), box[2:].tolist()
        color = map(int, color)
        image = cv2.rectangle(image, tuple(top_left), tuple(bottom_right),
                              tuple(color), line_thickness)
    return image


def get_img_rois(img,
                 boxes,
                 masks=None,
                 texts=None,
                 padding=100,
                 line_thickness=1,
                 font_size=0.5,
                 color=(255, 255, 255)):
    if not isinstance(img, np.ndarray):
        img = np.array(img)
    if len(img.shape) == 2:
        img = img[..., None]
    if img.shape[-1] == 1:
        img = img.repeat(3, -1)
    rois = np.array(boxes).round().astype(int)
    h, w = img.shape[:2]
    imgs_list = []
    if masks is None:
        masks = [None] * len(rois)
    if texts is None:
        texts = [""] * len(rois)
    img_info = np.iinfo(img.dtype)
    for roi, mask, text in zip(rois, masks, texts):
        x0, y0, x1, y1 = roi
        x0, x1 = np.clip([x0 - padding, x1 + padding], 0, w)
        y0, y1 = np.clip([y0 - padding, y1 + padding], 0, h)
        new_img = img[y0:y1, x0:x1, :].copy()

        if mask is not None:
            new_img = new_img.astype(float)
            mask = np.array(mask).squeeze()
            if mask.shape == img.shape[:2]:
                mask = mask[y0:y1, x0:x1, None] * color
                new_img = new_img + mask
            else:
                mask = cv2.resize(mask,
                                  (roi[2] + 1 - roi[0], roi[3] + 1 - roi[1]))
                mask = mask[..., None] * color
                mx0, mx1 = roi[0] - x0, roi[2] + 1 - x0
                ix0, ix1 = np.clip([mx0, mx1], 0, x1 - x0)
                my0, my1 = roi[1] - y0, roi[3] + 1 - y0
                iy0, iy1 = np.clip([my0, my1], 0, y1 - y0)
                new_img[iy0:iy1, ix0:ix1] += mask[(iy0 - my0):(iy1 - my0),
                                                  (ix0 - mx0):(ix1 - mx0)]
            new_img = new_img.clip(img_info.min,
                                   img_info.max).astype(img.dtype)

        if text:
            cv2.putText(new_img, str(text), (padding, padding),
                        cv2.FONT_HERSHEY_SIMPLEX, font_size, color,
                        line_thickness)
        imgs_list.append(new_img)

    return imgs_list


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# -*- coding: utf-8 -*-

import gzip
import json
import os
import os.path as osp
import pickle
import shutil
import tarfile
import time
from collections import Callable

import numpy as np
from PIL import Image

from .dicom import sitk, sitk_read_image

PARSER_EXT_DICT = {"txt": "txt",
                   "pickle": "pkl", "json": "json",
                   "torch": "pth", "sitk": ["dicom", "dcm", "nii"],
                   "image": ["png", "jpg", "jpeg", "bmp"]}


def _inverse_dict(d):
    inv_d = {}
    for k, v in d.items():
        if isinstance(v, list):
            inv_d.update({_v: k for _v in v})
        else:
            inv_d[v] = k
    return inv_d


EXT_TO_PARSER_DICT = _inverse_dict(PARSER_EXT_DICT)


def make_dir(*args):
    """
    the one-liner directory creator
    """
    path = osp.join(*[arg.strip(" ") for arg in args])
    if not osp.isdir(path):
        from random import random

        time.sleep(random() * 0.001)
        if not osp.isdir(path):
            os.makedirs(path)
    return path


def read_str(string, default_out=None):
    """
    the one-liner string parser
    """
    def invalid_entry(value):
        return value is None or value == ""

    def invalid_type(value):
        raise ValueError()

    if invalid_entry(string):
        return default_out

    if isinstance(string, str):
        parser = json.loads
    elif isinstance(string, bytes):
        parser = pickle.loads
    else:
        parser = invalid_type
    try:
        out = parser(string)
    except Exception:
        out = default_out
        print(string)
    return out


def load(filename, file_type="auto", **kwargs):
    """
    the one-liner loader
    Args:
        filename (str)
        file_type (str): support types: PARSER_EXT_DICT
    """

    if not isinstance(filename, str):
        return None

    if file_type == "auto":
        ext = filename.rpartition(".")[-1].lower()
        file_type = EXT_TO_PARSER_DICT.get(ext, "unknown")

    if file_type == "txt":
        with open(filename, "r") as f:
            result = f.readlines()
    elif file_type == "json":
        with open(filename, "r") as f:
            result = json.load(f, **kwargs)
    elif file_type == "pickle":
        with open(filename, "rb") as f:
            result = pickle.load(f, **kwargs)
    elif file_type == "torch":
        import torch
        result = torch.load(filename, map_location=torch.device("cpu"))
    elif file_type == "sitk":
        result = sitk_read_image(filename, **kwargs)
    elif file_type == "image":
        result = Image.open(filename)
        as_np = kwargs.get("as_np", False)
        if as_np:
            result = np.array(result)
    elif file_type == "unknown":
        raise ValueError(f"Unknown ext {filename}")
    else:
        raise NotImplementedError(f"Unknown file_type {file_type}")

    return result


def save(to_be_saved, filename, file_type="auto"):
    """
    the one-liner saver
    Args:
        to_be_saved (any obj)
        filename (str)
        file_type (str): support types: PARSER_EXT_DICT
    """
    if not isinstance(filename, str):
        return None

    if file_type == "auto":
        ext = filename.rpartition(".")[-1]
        file_type = EXT_TO_PARSER_DICT.get(ext, "unknown")

    if file_type == "txt":
        with open(filename, "w") as f:
            f.write(to_be_saved)
    elif file_type == "json":
        with open(filename, "w") as f:
            json.dump(to_be_saved, f)
    elif file_type == "pickle":
        with open(filename, "wb") as f:
            pickle.dump(to_be_saved, f)
    elif file_type == "torch":
        import torch
        torch.save(to_be_saved, filename)
    elif file_type == "sitk":
        sitk.WriteImage(to_be_saved, filename)
    elif file_type == "unknown":
        raise ValueError(f"Unknown ext {filename}")
    else:
        raise NotImplementedError(f"Unknown file_type {file_type}")


def recursive_copy(src, dst, softlink=False, overwrite=True, filter_fn=None):
    """
    recursively update dst root files with src root files
    Args:
        src (str): source root path
        dst (str): destination root path
        softlink (bool): Default False, if True, using os.symlink instead of copy
        overwrite (bool): Default True, if overwrite when file already exists
        filter_fn (function): given basename of src path, return True/False
    """
    src_file_list = []

    for root, dirs, files in os.walk(src):
        for filename in files:
            if isinstance(filter_fn, Callable) and not filter_fn(filename):
                continue
            relative_root = root.rpartition(src)[-1].lstrip("/ ")
            src_file_list.append((relative_root, filename))

    make_dir(dst)

    dst_file_list = []
    for root, dirs, files in os.walk(dst):
        for filename in files:
            relative_root = root.rpartition(src)[-1].lstrip("/ ")
            dst_file_list.append((relative_root, filename))

    for f in src_file_list:
        if f in dst_file_list:
            continue
        relative_root = f[0]
        make_dir(dst, relative_root)
        src_path = osp.join(src, *f)
        dst_path = osp.join(dst, *f)
        if osp.exists(dst_path):
            if overwrite:
                if osp.islink(dst_path):
                    if osp.isdir(dst_path):
                        os.rmdir(dst_path)
                    else:
                        os.unlink(dst_path)
                else:
                    os.remove(dst_path)
                print(f"{dst_path} exists, overwrite")
            else:
                print(f"{dst_path} exists, skip")
                continue
        if softlink:
            os.symlink(src_path, dst_path)
        else:
            shutil.copyfile(src_path, dst_path)


def unzip(src_path, dst_path):
    """
    the one-liner unzip function, currently support "gz" and "tgz"
    """
    if osp.isdir(dst_path):
        filename = osp.basename(src_path).rpartition(".")[0]
        dst_path = osp.join(dst_path, filename)

    if src_path.endswith(".gz"):
        with gzip.open(src_path, 'rb') as f_in:
            with open(dst_path, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

    elif src_path.endswith(".tgz"):
        with tarfile.open(src_path, "r:gz") as tar:
            tar.extractall(path=dst_path)

    else:
        raise NotImplementedError(f"unrecognized zip format {src_path}")

    return dst_path


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# -*- coding: utf-8 -*-


def list_to_str(float_list, decimal_count=3, with_bracket=True):
    """
    formatting list of floats
    """
    _string = ", ".join(["{:.%df}" % decimal_count] * len(float_list))
    if with_bracket:
        _string = "[" + _string + "]"
    return _string.format(*float_list)


def search_dir(module, key=""):
    """
    the one-liner dir
    """
    return [i for i in dir(module) if key in i.lower()]


__all__ = [k for k in globals().keys() if not k.startswith("_")]
